﻿// Copyright (C) Microsoft Corporation. All Rights Reserved.
// This code released under the terms of the Microsoft Public License
// (MS-PL, http://opensource.org/licenses/ms-pl.html).

using System;
using System.Globalization;
using System.Net;
using System.Threading;

/// <summary>
/// Class that demonstrates how to do a streaming HTTP PUT.
/// </summary>
/// <remarks>
/// The use of HttpListener requires this sample to be run with Administrator privileges.
/// The exception "HttpListenerException: Access is denied" means that hasn't been done.
/// </remarks>
static class StreamingHttpPut
{
    private static object _lock = new object();
    private static byte[] _data = new byte[5];
    private static string _uri = "http://localhost:80/";

    /// <summary>
    /// Entry point of sample.
    /// </summary>
    public static void Main()
    {
        try
        {
            // Create a simple HTTP listener
            using (var listener = new HttpListener())
            {
                listener.Prefixes.Add(_uri);
                listener.Start();

                // Queue a task to run the sample actions in sequence
                var actions = new Action[]
                {
                    PutWithWebClient,
                    PutWithNormalHttpWebRequest,
                    PutWithUnbufferedHttpWebRequest
                };
                ThreadPool.QueueUserWorkItem((state) =>
                    {
                        foreach (var action in actions)
                        {
                            Console.WriteLine();
                            Thread.Sleep(1000);
                            action();
                        }
                    });

                // Trivially handle each action's incoming request
                for (int i = 0; i < actions.Length; i++)
                {
                    var context = listener.GetContext();
                    var request = context.Request;
                    Log('S', "Got " + request.HttpMethod + " request");
                    using (var stream = request.InputStream)
                    {
                        while (-1 != stream.ReadByte())
                        {
                            Log('S', "Read request byte");
                        }
                        Log('S', "Request complete");
                    }
                    context.Response.Close();
                }
            }
        }
        catch (Exception e)
        {
            // Output diagnostic info about the failure
            Console.Error.WriteLine(e);
        }
    }

    /// <summary>
    /// Test action that uses WebClient's UploadData to do the PUT.
    /// </summary>
    private static void PutWithWebClient()
    {
        using (var client = new WebClient())
        {
            Log('C', "Start WebClient.UploadData");
            client.UploadData(_uri, "PUT", _data);
            Log('C', "End WebClient.UploadData");
        }
    }

    /// <summary>
    /// Test action that uses a normal HttpWebRequest to do the PUT.
    /// </summary>
    private static void PutWithNormalHttpWebRequest()
    {
        var request = (HttpWebRequest)(WebRequest.Create(_uri));
        request.Method = "PUT";
        Log('C', "Start normal HttpWebRequest");
        using (var stream = request.GetRequestStream())
        {
            foreach (var b in _data)
            {
                Thread.Sleep(1000);
                Log('C', "Writing byte");
                stream.WriteByte(b);
            }
        }
        Log('C', "End normal HttpWebRequest");
        ((IDisposable)(request.GetResponse())).Dispose();
    }

    /// <summary>
    /// Test action that uses an unbuffered HttpWebRequest to do the PUT.
    /// </summary>
    private static void PutWithUnbufferedHttpWebRequest()
    {
        var request = (HttpWebRequest)(WebRequest.Create(_uri));
        request.Method = "PUT";
        // Disable AllowWriteStreamBuffering allows the request bytes to send immediately
        request.AllowWriteStreamBuffering = false;
        // Doing nothing else will result in "ProtocolViolationException: When performing
        // a write operation with AllowWriteStreamBuffering set to false, you must either
        // set ContentLength to a non-negative number or set SendChunked to true.
        // The most widely supported approach is to set the ContentLength property
        request.ContentLength = _data.Length;
        Log('C', "Start unbuffered HttpWebRequest");
        using (var stream = request.GetRequestStream())
        {
            foreach (var b in _data)
            {
                Thread.Sleep(1000);
                Log('C', "Writing byte");
                stream.WriteByte(b);
            }
        }
        Log('C', "End unbuffered HttpWebRequest");
        ((IDisposable)(request.GetResponse())).Dispose();
    }

    /// <summary>
    /// Simple logging wrapper.
    /// </summary>
    /// <param name="role">Role of the caller.</param>
    /// <param name="message">Message to log.</param>
    private static void Log(char role, string message)
    {
        lock (_lock)
        {
            Console.WriteLine(
                DateTime.Now.ToString("HH:mm:ss.ff", CultureInfo.InvariantCulture) +
                " <" + role + "> " +
                message);
        }
    }
}
